@@ -5,17 +5,22 @@ from pytodotxt import Task
 
 
 class Searcher:
-    def __init__(self, text=None, casesensitive=True, default_threshold=None):
+    def __init__(self, text=None, casesensitive=True, default_threshold=None, hide_sequential=True):
         self.words = set()
         self.not_words = set()
         self.projects = set()
         self.not_projects = set()
         self.contexts = set()
         self.not_contexts = set()
+        self.ids = set()
+        self.not_ids = set()
+        self.after = set()
+        self.refs = set()
         self.done = None
         self.show_hidden = False
         self.priority = None
         self.default_threshold = default_threshold
+        self.hide_sequential = hide_sequential
         if default_threshold is None or len(default_threshold) == 0:
             self.default_threshold = 'today'
         self.threshold = self.default_threshold
@@ -23,6 +28,8 @@ class Searcher:
         self.created = None
         self.completed = None
         self.casesensitive = casesensitive
+        self.sources = None
+        self.task_by_id = None
 
         self.text = text
 
@@ -36,6 +43,10 @@ class Searcher:
         self.not_projects = set()
         self.contexts = set()
         self.not_contexts = set()
+        self.ids = set()
+        self.not_ids = set()
+        self.after = set()
+        self.refs = set()
         self.done = None
         self.show_hidden = False
         self.priority = None
@@ -44,6 +55,52 @@ class Searcher:
         self.created = None
         self.completed = None
 
+    def update_sources(self, sources):
+        self.sources = sources
+        self.task_by_id = {}
+        self.parents = {}
+        with_sequence = set()
+        for source in self.sources:
+            for task in source.tasks:
+                if len(task.attr_id) == 0:
+                    continue
+                if len(task.attr_after) > 0:
+                    with_sequence.add(task)
+                for taskid in task.attr_id:
+                    if taskid not in self.task_by_id:
+                        self.task_by_id[taskid] = set()
+                    self.task_by_id[taskid].add(task)
+
+        # for each task obtain a set of their direct and indirect parents
+        for task in with_sequence:
+            parents = self._parse_ids('after', task)
+            queue = parents.copy()
+            while len(queue) > 0:
+                otherid = queue.pop()
+                if otherid in self.parents:
+                    parents |= self.parents[otherid]
+                    break
+                parents.add(otherid)
+
+                if otherid not in self.task_by_id:
+                    continue
+                for other in self.task_by_id[otherid]:
+                    queue |= self._parse_ids('after', other).difference(parents)
+            if len(parents) > 0:
+                for taskid in task.attr_id:
+                    if taskid not in self.parents:
+                        self.parents[taskid] = set()
+                    self.parents[taskid] |= parents
+
+    def _parse_ids(self, keyword, task):
+        these = set()
+        for value in task.attributes.get(keyword, []):
+            these |= {that for that in value.split(',')}
+        if self.casesensitive:
+            return {that.lower() for that in these}
+        else:
+            return these
+
     def parse(self):
         self.reset()
         text = self.text
@@ -159,6 +216,41 @@ class Searcher:
                     self.threshold = None
                 else:
                     self.threshold = get_relative_date(value, Task.DATE_FMT) or value
+
+            elif part.startswith('id:'):
+                _, value = part.split(':', 1)
+                if len(value) > 0:
+                    value = {that for that in value.split(',')}
+
+                    if do_not:
+                        self.not_ids |= value
+                    else:
+                        self.ids |= value
+                else:
+                    if do_not:
+                        self.not_words.add('id:')
+                    else:
+                        self.words.add('id:')
+
+            elif part.startswith('after:') and self.after is not None:
+                _, value = part.split(':', 1)
+                if len(value) == 0:
+                    self.after = None
+                else:
+                    values = set(value.split(','))
+                    if do_not:
+                        pass
+                    else:
+                        self.after |= values
+
+            elif part.startswith('ref:') and self.refs is not None:
+                _, value = part.split(':', 1)
+                if len(value) > 0:
+                    values = set(value.split(','))
+                    if do_not:
+                        pass
+                    else:
+                        self.refs |= values
             
             else:
                 if do_not:
@@ -175,6 +267,9 @@ class Searcher:
                     self.match_done(task),
                     self.match_hidden(attrs),
                     self.match_priority(task),
+                    self.match_ids(task),
+                    self.match_refs(task),
+                    self.match_after(task, attrs),
                     self.match_due(attrs),
                     self.match_created(task),
                     self.match_completed(task),
@@ -284,6 +379,50 @@ class Searcher:
 
         return completed[0] < task_completed < completed[1]
 
+    def match_ids(self, task):
+        if len(self.not_ids) == 0 and len(self.ids) == 0:
+            return True
+
+        ids = self._parse_ids('id', task)
+
+        return (len(self.ids) == 0 or len(self.ids.intersection(ids)) > 0) and \
+               (len(self.not_ids) == 0 or len(self.not_ids.intersection(ids)) == 0)
+
+    def match_refs(self, task):
+        if len(self.refs) == 0:
+            return True
+
+        ids = self._parse_ids('after', task)
+        ids |= self._parse_ids('ref', task)
+
+        return len(ids) > 0 and \
+               len(self.refs.intersection(ids)) > 0
+
+    def match_after(self, task, attrs):
+        if self.after is None:
+            return True
+
+        if len(self.after) == 0 and \
+           len(attrs.get('after', [])) == 0:
+            return True
+
+        if self.task_by_id is None:
+            return True
+
+        parents = self._parse_ids('after', task)
+
+        if len(self.after) == 0:
+            # normal sequence behaviour
+            return not self.hide_sequential or \
+                   all([parentid not in self.task_by_id or
+                        any([other.is_completed for other in self.task_by_id[parentid]]) 
+                        for parentid in parents])
+        else:
+            # collect all indirect parents, too
+            for taskid in task.attr_id:
+                parents |= self.parents.get(taskid, set())
+            return any([parent in parents for parent in self.after])
+
 
 def get_relative_date(text, fmt=None):
     weekdays = ['mon', 'tue', 'wed', 'thu', 'fri',
